# Copyright (c) 2004 by RUNYAGA, LLC.
# http://www.runyaga.com/

"""
$Id: api.py 758 2005-06-28 23:02:39Z sidnei $
"""
from itertools import ifilter
import re
from datetime import date, datetime
from zope.app import zapi
from zope.interface import directlyProvides, directlyProvidedBy
from zope.component import ComponentLookupError
from zope.app.container.interfaces import IContained, IReadContainer
from zope.app.container.contained import ContainedProxy
from zope.app.location.interfaces import ILocation
from zope.app.location.location import LocationProxy
from zope.proxy import sameProxiedObjects
from zope.security import checkPermission
from zope.app.security.permission import checkPermission as checkValidPermission
from zope.app.rdb import queryForResults
from sip.schema.interfaces import IMutableSchema
from zope.proxy import removeAllProxies
from sip.interfaces.csv import IProcessor, ITempStorage, \
     IMassager, ILookupHelper
from sqlos.interfaces import IISQLObject
from zope.event import notify
from zope.app.event.objectevent import ObjectCreatedEvent, ObjectModifiedEvent

def factory(name, context=None):
    return zapi.getUtility(IISQLObject, name, context=context)

def queryService(name, context=None, default=None):
    try:
        return zapi.getService(name, context=context)
    except ComponentLookupError:
        return default

def _queryContextByIface(context, iface):
    impls = iface.providedBy
    parent = context
    while not impls(parent) and parent is not None:
        if hasattr(parent, 'context') and impls(parent.context):
            return parent.context
        parent = zapi.getParent(parent)
    return parent

def _getContextByIface(context, iface):
    obj = _queryContextByIface(context, iface)
    if obj is None:
        raise ComponentLookupError, ("Couldn't find object implementing "
                                     "%r in context %r" % (iface, context))
    return obj

def queryContainer(context):
    from sip.interfaces.container import IContainer
    from sip.interfaces.location import ILocation
    # had to check for ILocation so that filtering of ContainerTypes
    # worked when adding a Container to a Location
    result = _queryContextByIface(context, IContainer) \
             or _queryContextByIface(context, ILocation)
    return result

def queryProject(context):
    from sip.interfaces.project import IProject
    return _queryContextByIface(context, IProject)

def queryRoot(context):
    from sip.interfaces import IAppRoot
    return _queryContextByIface(context, IAppRoot)

def querySample(context):
    from sip.interfaces.sample import ISample
    return _queryContextByIface(context, ISample)

def getContainer(context):
    from sip.interfaces.container import IContainer
    return _getContextByIface(context, IContainer)

def getProject(context, appRoot=None):
    from sip.interfaces.project import IProject
    try:
        return _getContextByIface(context, IProject)
    except ComponentLookupError:
        if IContained.providedBy(context):
            raise
    if appRoot is None:
        raise TypeError, 'Must provide a contained context OR an application root'
    elif not hasattr(context, 'project_id'):
        raise TypeError, 'Must provide a contained context OR one that has a project_id'
    else:
        projects = appRoot['projects']
        for project in projects.values():
            if project.id == context.project_id:
                return project
        raise ComponentLookupError, ("Couldn't find project "
                                     "%r in context %r" % (context.project_id, context))

def getRoot(context):
    from sip.interfaces import IAppRoot
    return _getContextByIface(context, IAppRoot)

def getSample(context):
    from sip.interfaces.sample import ISample
    return _getContextByIface(context, ISample)

def getSchemaFor(type_id, context):
    fact = factory('sip.SampleTypeSchema', context=context)
    project = getProject(context)
    results = fact.select(('sample_type_id = %s and '
                              'project_id = %s' % (type_id, project.id)))
    if results.count():
        schema = results[0]
    else:
        schema = fact(sample_type_id=type_id,
                         project_id=project.id)
    project = IReadContainer(project)
    sts = project['schemas']['sip.SampleTypeSchema.%s' % schema.id]
    return IMutableSchema(sts)

def getAliquotsForShipment(shipment):
    ct = factory('sip.Container', context=shipment)
    # Root must always exist. If it doesn't, theres an error.
    root = ct.select(
        'container_shipment.shipment_id = %d and '
        'container_shipment.container_id = container.container_id and '
        'container.parent_id = 1' % shipment.id,
        clauseTables=['container_shipment'])[0]
    aliquots = getAliquotsForContainer(root)
    for container in recurseGetContainers(root):
        aliquots.extend(getAliquotsForContainer(container))
    return aliquots

def recurseGetContainers(container):
    ct = factory('sip.Container', context=container)
    containers = list(
        ct.select('parent_id = %d' % container.id))
    for c in containers[:]:
        containers.extend(recurseGetContainers(c))
    return containers

def getAliquotsForContainer(container):
    al = factory('sip.Aliquot', context=container)
    return list(al.select('container_id = %d' % container.id))

def uniqueId(context):
    from sip.interfaces import IAppRoot
    parent = zapi.getParent(context)
    url = zapi.getName(context)
    while not IAppRoot.providedBy(context) and parent is not None:
        url = "%s/%s" % (zapi.getName(parent), url)
        context = parent
        parent = zapi.getParent(context)
    return url

def makeName(obj):
    name = '%s.%s' % (obj.__class__.__name__, obj.id)
    return name

def findParent(obj, appRoot):
    if not hasattr(obj, 'project_id'):
        raise ValueError, ('To find the parent of an object, it must be '
                           'an SQLObject and have a project_id field')
    class_name = obj.__class__.__name__
    return zapi.traverse(appRoot, 'projects/sip.Project.%s/%ss'
                         % (obj.project_id, class_name.lower()))

def contained(obj, parent=None, name=None, appRoot=None):
    if (parent is None) and (appRoot is None):
        raise TypeError, 'Must provide either a parent or an application root'

    if not IContained.providedBy(obj):
        if ILocation.providedBy(obj):
            directlyProvides(obj, IContained, directlyProvidedBy(obj))
        else:
            obj = ContainedProxy(obj)

    oldparent = obj.__parent__
    oldname = obj.__name__

    if (oldparent is None) or not (oldparent is parent
                                   or sameProxiedObjects(oldparent, parent)):
        if parent is None:
            parent = findParent(obj, appRoot)
        obj.__parent__ = parent

    if oldname != name and name is not None:
        obj.__name__ = name

    return obj

def containedItems(items, container):
    c = lambda item, container=container: \
        (item[0], contained(item[1], container, item[0]))
    return map(c, items)

def uncontained(item):
    if IContained.providedBy(item):
        item.__parent__ = None
        item.__name__ = None
    return item

def uncontainedItems(items):
    return map(uncontained, items)

def canView(context):
    perm_id = 'sip.View%s' % context.__class__.__name__
    checkValidPermission(context, perm_id)
    allow = checkPermission(perm_id, context)
    return allow

def itemsCanView(items, context, factory):
    if not items: return []
    perm_id = 'sip.View%s' % factory.__name__
    checkValidPermission(context, perm_id)
    cp = checkPermission
    c = lambda item, perm_id=perm_id: cp(perm_id, item[1])
    return ifilter(c, items)

def checkSIPPermission(context, perm_id):
    checkValidPermission(context, perm_id)
    allow = checkPermission(perm_id, context)
    return allow

def getFieldId(context, sample_id, fname, stname=''):
    fact = factory('sip.Sample', context=context) # just need the connection
    conn = fact._connection._connection
    query = "select get_field_id(%d,'%s','%s') as id" % (sample_id, fname,
                                                           stname)
    #query = """
    #declare @status int
    #exec @status = get_field_id %d, '%s', '%s'
    #select id = @status"""  % (sample_id, fname, stname)

    #query = "exec get_field_id %d, '%s', '%s'" % (sample_id, fname, stname)

    cursor = conn.conn.cursor()
    #cursor.callproc('get_field_id', [sample_id, str(fname), str(stname)])
    #results = cursor.fetchone()
    #id = results[0]
    results = queryForResults(conn, query)
    id = results[0].id
    id = int(id)
    return id

def getFields(ctx, project_id, sample_type_id):
    fact = factory('sip.Sample', context=ctx) # just need the connection
    conn = fact._connection._connection
    query = "select * from schema_field where schema_id = (select " +\
            "sample_type_schema_id from sample_type_schema where " +\
            "project_id=%d and sample_type_id=%d)"%(project_id, sample_type_id)
    results = queryForResults(conn, query)
    return results

import thread
_mutex = thread.allocate_lock()

def allocateAliquotIds(ctx, count=1):
    fact = factory('sip.Aliquot', context=ctx)
    conn = fact._connection._connection
    cursor = conn.conn.cursor()
    _mutex.acquire()
    query = 'select * from get_aliquot_ids(%d) as id' % count
    try:
        results = queryForResults(conn, query)
    finally:
        _mutex.release()
    return [r.id for r in results]

tname_match = r"""template: *([a-zA-Z0-9._]+)"""
def lookupProcessor(data=None, template=None):
    if template is None and data is None:
        raise ValueError, 'Neither data nor template found, Unable to proceed'
    if template is None:
        match = re.search(tname_match, data)
        if not match:
            raise ValueError, 'Template section not found. Unable to proceed'
        template = match.group(1).strip()
    utilities = zapi.getService(zapi.servicenames.Utilities)
    proc = utilities.queryUtility(IProcessor, name=template)
    if proc is None:
        raise ComponentLookupError, 'Processor not found for %s' % template
    return proc, template

def lookupStorage(template):
    utilities = zapi.getService(zapi.servicenames.Utilities)
    proc = utilities.queryUtility(ITempStorage, name=template)
    if proc is None:
        proc = utilities.getUtility(ITempStorage, name='default')
    return proc

def lookupMassager(template):
    utilities = zapi.getService(zapi.servicenames.Utilities)
    proc = utilities.queryUtility(IMassager, name=template)
    if proc is None:
        proc = utilities.getUtility(IMassager, name='default')
    return proc

def getLookupUtility(name):
    utilities = zapi.getService(zapi.servicenames.Utilities)
    util = utilities.queryUtility(ILookupHelper, name=name)
    return util

def createTID(template):
    from random import randint
    # XXX Should be replaced by some persistent
    # id generator.
    return randint(1, 100000)

def getSubjectType(context, id):
    fact = factory('sip.SubjectType', context=context)
    return fact.get(int(id)).name

def getContainerPathFromId(context, id):
    fact = factory('sip.Container', context=context)
    container = fact.get(id)
    titles = []
        # The root container of a location has
        # parent_id == 1, which is the root container,
        # used for referential integrity
    while (container.parent_id and
           int(container.parent_id) != 1):
        titles.append(container.name)
        container = fact.get(container.parent_id)
    location = getLocation(context, container)
    titles.append(location.name)
    titles.reverse()
    path = 'path://' + '/'.join(titles)
    return path

def convertFieldAndValue(field, value):
    """ Since we are using strings to store all the values
        we must cast the field and value into something sane.
    """
    field_type = removeAllProxies(field)._type

    if value == '' or value == 'None':
        return field_type, None

    if field_type is not None:
        if field_type is bool and type(value) is not bool:
            # We expect all bools to be text from database
            if value[:1].lower() in ('0', 'f'):
                value=False
            else:
                value=True
        if field_type in (datetime, date) and type(value) not in (datetime, date):
            if value:
                #XXX This smacks of some sort of policy that is tied
                #    to the database implementation AS WELL AS the locale.
                #    Please refactor this and put a test in.
                sep=None
                for sep in ['/', '-']:
                    if value.find(sep)!=-1:
                        break
                if sep is None:
                    raise ValueError, "Could not find token to separate date"
                value = value[:10] # lets make sure we get rid of time if any
                yyyy,mm,dd = value.split(sep)
                try:
                    value = date(int(yyyy), int(mm), int(dd))
                except ValueError:
                    if len(dd)==2:
                        if int(dd)>90: dd='19%s' % dd
                        else: dd='20%s' % dd
                    value = date(int(dd), int(yyyy), int(mm))
                field_type = date

        # In Zope the field_type for a int is a tuple (int, long)
        if field_type == (int, long) and type(value) not in (int, long):
            if value:
                value=int(value)
        if field_type is float and type(value) is not float:
            if value:
                value=float(value)
        if field_type is unicode and type(value) is not unicode:
            if value is not None:
                value=unicode(value)

    return field_type, value


def getLocation(context, container):
    fact = factory('sip.ContainerLocation', context=context)
    project = IReadContainer(getRoot(context))
    locations = project['locations']
    results = fact.selectBy(container_id=container.id)
    results = [r for r in results]
    location = results[0].location_
    contained(location, locations, 'sip.Location.%d' % location.id)
    return location

def setLocation(context, container, location):
    fact = factory('sip.ContainerLocation', context=context)
    results = fact.selectBy(container_id=container.id)
    results = [r for r in results]
    if len(results):
        res = results[0]
        res.location_id = location.id
        notify(ObjectModifiedEvent(res))
    else:
        kw = {'container_id': container.id,
              'location_id': location.id}
        new = fact(**kw)
        notify(ObjectCreatedEvent(new))
