# Copyright (c) 2004 by RUNYAGA, LLC.
# http://www.runyaga.com/


"""
$Id: test_csv.py 758 2005-06-28 23:02:39Z sidnei $
"""

import unittest
from os import curdir
from os.path import join, abspath, dirname
import difflib
from zope.app import zapi
from zope.schema import Text
from zope.app.container.interfaces import IReadContainer

from sip.browser import api
from sip.csv.ftests.base import Base as BaseTest
from sip.csv.processor import *
from sip.csv.storage.exceptions import StorageError
from sip.interfaces.adapter import IAugmentedSample,\
             ISampleWithAliquots, ISampleWithOriginAliquots
from sip.csv.lookup import getContainerId

try:
    __file__
except NameError:
    # Called directly, so no __file__ global exists.
    _prefix = abspath(curdir)
else:
    # Called by another script.
    _prefix = abspath(dirname(__file__))

class Test(BaseTest):
    context_attr = 'project'
    level = 3
    prefix = ''

    def setUp(self):
        super(Test, self).setUp()
        self.project = self._addProject(name='Foo')
        self.project2 = self._addProject(name='Bar')
        self.technician = self._addTechnician(username='scole')
        self._addWeightUnit(label='g')
        self._addVolumeUnit(label='mL')
        self._addConcentrationUnit(label='g/mL')
        self.foo = self._addSampleType(self.project, name='SampleType Foo')
        schema = api.getSchemaFor(self.foo.id, self.project)
        field = Text(title=u'HOL', required=False)
        schema.addField(u'hol', field)
        schema.constrain(u'hol', u'python:value > 0')
        self.bar = self._addSampleType(self.project, name='SampleType Bar')
        schema = api.getSchemaFor(self.bar.id, self.project)
        field = Text(title=u'Diluient', required=False)
        schema.addField(u'diluient', field)
        ctype = self._addContainerType()
        self.location = self._addLocation(type_id=ctype.id, name=u'location')
        self.container = self._addContainer(self.location, ctype.id,
            name=u'container')
        self.subcontainer0 = self._addContainer(self.container, ctype.id,
            name=u'subcontainer0')
        self.subcontainer1 = self._addContainer(self.container, ctype.id,
            name=u'subcontainer1')
        self.subcontainer2 = self._addContainer(self.container, ctype.id,
            name=u'subcontainer2')
        self.location2 = self._addLocation(type_id=ctype.id, name=u'location2')
        self.container1 = self._addContainer(self.location2, ctype.id,
            name=u'container1')
        self.subcontainerA = self._addContainer(self.container1, ctype.id,
            name=u'subcontainerA')
        self.shipment1 = self._addShipment(self.project, name='ShipmentZ')
        self.containerz = self._addContainer(self.shipment1, ctype.id,
            name=u'subcontainerz')

    def test_process_csv(self):
        context = getattr(self, self.context_attr)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context, request=self.request)
        # Make sure a processor called with no ticket id (tid)
        # creates a new tid
        self.failIf(tid is None)
        # And that some log is created
        self.failUnless(log)
        # Needs to run setup again now, because the transaction
        # was aborted. IIRC!
        self.setUp()
        context = getattr(self, self.context_attr)
        # Compare value stored under tid with the
        # massager's output to make sure data is sane
        stored = str(proc.storage.load(context=context, tid=tid))
        massaged = str(proc.massager(context=context, data=self.data))
        self.failUnless(massaged == stored,
                        ("%s\n%s\n%s" %
                         (self.fname, massaged, stored)))
        # Then compare massager output with some hand-made data
        # to make sure the massager is doing sane stuff
        massaged = str(proc.massager(context=context, data=self.data))
        differ = difflib.Differ()
        diff = '\n'.join(list(differ.compare(self.stripped.splitlines(),
                                             massaged.splitlines())))
        self.failUnless(massaged == self.stripped.strip(),
                        ("%s\n%s" %
                         (self.fname, diff)))
        # Now run a real process, with the previously
        # generated tid
        log, n_tid = proc(data=self.data, context=context,
                          tid=tid, dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # Which should return a None tid
            self.failUnless(n_tid is None)
            # And expire the data for the previous
            # tid from the storage
            self.assertRaises(StorageError, proc.storage.load, None, tid)
            # And still generate some log
            self.failUnless(log)
        except:
            print 'log:\n' + log
            raise

    def skeleton_test(self):
        "You can copy and modify this to write a new test"
        context = self.project
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None)
            # But some log is created
            self.failUnless(log)
        except:
            print 'log:\n' + log
            raise

class Test_AddSampleAliquot(Test):
    fname = 'add_sample_aliquot.csv'
    prefix = ''
    proc_class = AddSampleAliquot

    def test_CreateSamples(self):
        context = self.project
        samples = zapi.traverse(context, 'samples')
        nsamples = len(samples)
        aliquots = zapi.traverse(context, 'aliquots')
        naliquots = len(aliquots)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None)
            # But some log is created
            self.failUnless(log)
            self.failUnless(len(samples) == nsamples + 4,
                            'created %s samples, expected 4'
                            % (len(samples) - nsamples))
            self.failUnless(len(aliquots) == naliquots + 6,
                            'created %s aliquots, expected 6'
                            % (len(aliquots) - naliquots))
            found = False
            for sample in samples.values():
                if sample.subject_id == '46193':
                    found = True
                    self.failUnless(sample.name == str(sample.id))
                    break

            if not found:
                self.fail("Couldn't find sample $SAMPLE1")
        except:
            print 'log:\n' + log
            raise

    def test_DumpSample(self):
        context = self.project
        proc = api.lookupProcessor(self.data, self.fname)[0]
        samples = zapi.traverse(context, 'samples')
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        for sample in samples.values():
            # XXX Need to add comparison against handmade data
            # to make sure we get the expected output.
            result = zapi.getView(sample, 'dump.csv', self.request)()

def _addSample_force(context, name, id, stype, subtype, other_id=None):
    factory = api.factory('sip.Sample', context=context)
    other_id = other_id or u''
    factory(id = id,
            name = name,
            type_id = stype.id,
            project_id=context.id,
            subject_id = u'1',
            other_id = other_id,
            technician_id = None,
            comments = u'',
            subject_type_id = subtype.id,
            commit_message = u'test',
            collection_date = u'1/1/1970')

def _addAliquot_force(context, sample_id, id):
    factory = api.factory('sip.Aliquot', context=context)
    factory(id = id,
            project_id=context.id,
            sample_id = sample_id,
            volume_id = None,
            concentration_id = None,
            weight_id = None,
            container_id = None,
            volume = None,
            concentration = None,
            weight = None,
            commit_message = u'test')

class Test_ModifyAliquot(Test):
    fname = 'modify_aliquot.csv'
    proc_class = ModifyAliquot
    context_attr = 'sip'

    def setUp(self):
        super(Test_ModifyAliquot, self).setUp()
        _addSample_force(context=self.project, id=12345,
                         name='Sample A', stype=self.foo, subtype=self.animal)
        _addSample_force(context=self.project, id=23456,
                          name='Sample B', stype=self.bar, subtype=self.animal)
        _addSample_force(context=self.project2, id=34456,
                          name='Sample C', stype=self.foo, subtype=self.animal)
        _addAliquot_force(context=self.project, sample_id=12345, id=23412)
        _addAliquot_force(context=self.project, sample_id=23456, id=12311)
        _addAliquot_force(context=self.project2, sample_id=34456, id=41422)

    def test_ModifySamplesAliquots(self):
        context = self.sip
        sample_factory = api.factory('sip.Sample', context=context)
        nsamples = sample_factory.select().count()
        assert nsamples == 3, nsamples
        aliquot_factory = api.factory('sip.Aliquot', context=context)
        naliquots = aliquot_factory.select().count()
        assert naliquots == 3, naliquots
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None, log)
            # But some log is created
            self.failUnless(log)
            nsamples = sample_factory.select().count()
            naliquots = aliquot_factory.select().count()
            self.failUnless(nsamples == 3,
                            'have %s samples, expected 3'
                            % nsamples)
            self.failUnless(naliquots == 4,
                            'have %s aliquots, expected 4'
                            % naliquots)
            sample_a = sample_factory.get(12345)
            sample_a = api.contained(sample_a, appRoot=self.sip)
            self.failUnless(sample_a.name == u'Sample A')
            self.failUnless(
                sample_a.subject_id == u'80629',
                'Sample was not edited - subject_id is %r, should be 80629'
                % sample_a.subject_id)
            sample_a = IAugmentedSample(sample_a)
            self.failUnless(
                sample_a.hol == u'4',
                'Sample additional data was not edited - HOL is %r, should be 4'
                % sample_a.hol)
            aliquot_a = aliquot_factory.get(23412)
            aliquot_a = api.contained(aliquot_a, appRoot=self.sip)
            self.failUnless(
                aliquot_a.sample_id == sample_a.id,
                aliquot_a.sample_id)
            self.failUnless(
                aliquot_a.weight == 1.0,
                'Aliquot was not edited - weight is %r, should be 1.0'
                % aliquot_a.weight)
        except:
            print 'log:\n' + log
            raise

    def test_ModifySamplesAliquotsNoUpdate(self):
        context = self.sip
        sample_factory = api.factory('sip.Sample', context=context)
        nsamples = sample_factory.select().count()
        assert nsamples == 3, nsamples
        aliquot_factory = api.factory('sip.Aliquot', context=context)
        naliquots = aliquot_factory.select().count()
        assert naliquots == 3, naliquots
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        sample_b = sample_factory.get(23456)
        aliquot_b = aliquot_factory.get(12311)
        old_sample_message = sample_b.commit_message
        old_aliquot_message = aliquot_b.commit_message
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None, log)
            # But some log is created
            self.failUnless(log)
            self.failUnless(sample_b.commit_message == old_sample_message)
            self.failUnless(aliquot_b.commit_message == old_aliquot_message)
        except:
            print 'log:\n' + log
            raise

class RModifyAliquot(ModifyAliquot):

    def load_data(self, **kw):
        data = super(ModifyAliquot, self).load_data(**kw)
        samples = data.tables['sample']
        aliquots = data.tables['aliquots']
        ad = data.tables['additional_data']
        samples.update(0, 'subject_id', u'80629')
        samples.update(0, 'other_id', u'13234')
        ad.update(0, 'field_value', 4)
        aliquots.update(0, 'weight', 1.0)
        aliquots.update(0, 'weight_id', 2)
        return data


class TestRoundtripModifyAliquot(Test):
    fname = 'modify_aliquot.csv'
    prefix = 'r_'
    proc_class = RModifyAliquot
    context_attr = 'sip'

    def setUp(self):
        super(TestRoundtripModifyAliquot, self).setUp()
        _addSample_force(context=self.project, id=12345,
                         name='Sample A', stype=self.foo,
                         subtype=self.animal,
                         other_id=u'12234')
        _addAliquot_force(context=self.project, sample_id=12345, id=23412)
        sample = zapi.traverse(self.project, 'samples/sip.Sample.12345')
        self.data = zapi.getView(sample, 'dump.csv', self.request)()

    def test_ModifySamplesAliquots(self):
        context = self.sip
        sample_factory = api.factory('sip.Sample', context=context)
        nsamples = sample_factory.select().count()
        assert nsamples == 1, nsamples
        aliquot_factory = api.factory('sip.Aliquot', context=context)
        naliquots = aliquot_factory.select().count()
        assert naliquots == 1, naliquots
        # Hardcoded processor.
        proc = self.proc_class(self.fname)
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None, log)
            # But some log is created
            self.failUnless(log)
            nsamples = sample_factory.select().count()
            naliquots = aliquot_factory.select().count()
            self.failUnless(nsamples == 1,
                            'have %s samples, expected 1'
                            % nsamples)
            self.failUnless(naliquots == 1,
                            'have %s aliquots, expected 1'
                            % naliquots)
            sample_a = sample_factory.get(12345)
            sample_a = api.contained(sample_a, appRoot=self.sip)
            self.failUnless(sample_a.name == u'Sample A')
            self.failUnless(
                sample_a.subject_id == u'80629',
                'Sample was not edited - subject_id is %r, should be 80629'
                % sample_a.subject_id)
            self.failUnless(
                sample_a.other_id == u'13234',
                'Sample was not edited - other_id is %r, should be 13234'
                % sample_a.subject_id)
            sample_a = IAugmentedSample(sample_a)
            self.failUnless(
                sample_a.hol == u'4',
                'Sample additional data was not edited - HOL is %r, should be 4'
                % sample_a.hol)
            aliquot_a = aliquot_factory.get(23412)
            aliquot_a = api.contained(aliquot_a, appRoot=self.sip)
            self.failUnless(
                aliquot_a.sample_id == sample_a.id,
                aliquot_a.sample_id)
            self.failUnless(
                aliquot_a.weight == 1.0,
                'Aliquot was not edited - weight is %r, should be 1.0'
                % aliquot_a.weight)
        except:
            print 'log:\n' + log
            raise

class Test_AddContainer(BaseTest):
    fname = 'add_containers.csv'
    proc_class = AddContainer
    prefix = ''
    level = 3

    def setUp(self):
        super(Test_AddContainer, self).setUp()
        project = self._addProject(name=u'Genetics:COR')
        self.project = project
        root_container = self._addContainerType(name=u'Freezer')
        self._addContainerType(name=u'master rack', root=False)
        self._addContainerType(name=u'rack', root=False)
        self._addContainerType(name=u'box', root=False)
        self._addContainerType(name=u'cell', root=False)
        location = self._addLocation(root_container.id,
                                     name=u'freezer name mr')
        project.valid_locations = [location.id]


    def test_CreateContainers(self):
        context = self.project
        locations = zapi.traverse(context, 'locations')
        locations = [l for l in locations.values()]
        self.failUnless(len(locations) == 1)
        location = locations[0]
        self.failUnless(location.name == 'freezer name mr')
        adapted = IReadContainer(location)
        self.failUnless(len(adapted) == 0)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None)
            self.failUnless(log)
            self.failUnless(len(adapted) == 2)
            paths = ['path://freezer name mr/FooContainer',
                     'path://freezer name mr/FooContainer/rack',
                     'path://freezer name mr/BarContainer']
            for x in range(1,11):
                for y in range(1,11):
                    paths.append(
                      'path://freezer name mr/BarContainer/cell-%dx%d' % (x,y))
            for x in range(1,6):
                for y in range(1,6):
                    paths.append(
                      'path://freezer name mr/BarContainer/cell-1x3/sub-%dx%d'
                      % (x,y))
            for path in paths:
                self.failUnless(getContainerId(context, path) is not None)

        except:
            print 'log:\n' + log
            raise


class Test_SplitAliquot(Test):
    """ We will create a Sample and an aliquot then split
        the aliquot into 3 more aliqouts.  The result should
        be 1 sample with 4 aliquots. """

    fname = 'split_aliquot.csv'
    proc_class = SplitAliquot
    _test_sample_id = 207152

    def setUp(self):
        super(Test_SplitAliquot, self).setUp()
        _addSample_force(context=self.project, id=self._test_sample_id,
                         name='Sample Split', stype=self.foo,
                         subtype=self.animal)
        _addAliquot_force(context=self.project,
                          sample_id=self._test_sample_id, id=54321)

    def test_SplitAliquot(self):
        context = self.project
        sample = zapi.traverse(context, ('samples/sip.Sample.%s' %
                                         self._test_sample_id))
        adapted_sample = ISampleWithAliquots(sample)
        aliq_count = len(adapted_sample.aliquots)
        adapted_container = IReadContainer(self.containerz)
        aliquot_count_container = len(adapted_container)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            self.failUnless(tid is None)
            self.failUnless(log)
            # Now lets readapt the sample to get the new aliquots.
            adapted_sample = ISampleWithAliquots(sample)
            split_count = len(adapted_sample.aliquots)
            self.failUnless(split_count == aliq_count + 4,
                            'created %s aliquots expected 4'
                            % (split_count-aliq_count))
            adapted_container = IReadContainer(self.containerz)
            aliquot_count_container += 1
            self.assertEquals(len(adapted_container), aliquot_count_container)
        except:
            print 'log:\n' + log
            raise


class Test_ModifyContainer(Test):
    fname = 'modify_containers.csv'
    proc_class = ModifyContainer

    def test_ReparentContainer(self):
        context = self.project
        container_factory = api.factory('sip.Container', context=context)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        # Make sure we got a processor
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            # If all is ok, proc() expires the tid, so it is None
            self.failUnless(tid is None, log)
            # But some log is created
            self.failUnless(log)
            self.subcontainer0.sync()
            self.subcontainer1.sync()
            self.subcontainer2.sync()
            self.subcontainerA.sync()
            self.failUnless(self.subcontainer0.parent_id == self.container1.id)
            self.failUnless(
                self.subcontainer1.parent_id == self.subcontainerA.id)
            shipment_root = self.shipment1.getRootContainer()
            self.failUnless(
                self.subcontainerA.parent_id == shipment_root.id)
            self.failUnless(self.subcontainer2.parent_id == self.container.id)
        except:
            print 'log:\n' + log
            raise

class Test_DeriveSample(BaseTest):
    fname = 'derive_sample.csv'
    proc_class = DeriveSample
    prefix = ''
    level = 3

    def setUp(self):
        super(Test_DeriveSample, self).setUp()
        project = self._addProject(name=u'Genetics: COR')
        self.project = project
        root_container = self._addContainerType(name=u'Freezer1')
        fct = self._addContainerType(name=u'FreezerContainer', root=False)
        fsct = self._addContainerType(name=u'FreezerSubContainer', root=False)
        location = self._addLocation(root_container.id, name=u'Freezer1')
        location_id = location.id
        project.valid_locations = [location_id]
        parent_id = getContainerId(project, 'path://Freezer1')
        fc = self._addContainer(location, fct.id, name='container')
        fc.name = 'container'
        sub_container = self._addContainer(fc, fsct.id,
                                 name='subcontainer0')
        sub_container.name = 'subcontainer0'
        self.container = sub_container
        st = self._addSampleType(project, name = 'Monkey Blood')
        schema = api.getSchemaFor(st.id, self.project)
        field = Text(title=u'HOL', required=False)
        schema.addField(u'hol', field)
        sample1 = _addSample_force(project, 'Sample A', 12345, st,
                                   subtype=self.animal)
        sample2 = _addSample_force(project, 'Sample B', 23456, st,
                                   subtype=self.animal)

        _addAliquot_force(project, 12345, 12345)
        _addAliquot_force(project, 23456, 23456)
        self._addWeightUnit(exponent = 1, label = 'g')
        self._addVolumeUnit(exponent = 1, label = 'mL')


    def test_DeriveSample(self):
        context = self.project
        aliquot_factory = api.factory('sip.Aliquot', context=context)
        sample_factory = api.factory('sip.Sample', context=context)
        proc = api.lookupProcessor(self.data, self.fname)[0]
        self.failIf(proc is None)
        log, tid = proc(data=self.data, context=context,
                        dry_run=False, request=self.request)
        log = '\n  '.join(log.splitlines()).strip()
        try:
            self.failUnless(tid is None, log)
            self.failUnless(log)
            sample = sample_factory.select("name='Sample C'")
            self.failUnless(sample.count() == 1)
            sample = sample[0]
            sample = zapi.traverse(context, 'samples/sip.Sample.%s' % sample.id)
            adapted = ISampleWithOriginAliquots(sample)
            origin_aliquots = adapted.origin_aliquots.values()
            self.failUnless(len(origin_aliquots) == 2)
            oas = {}
            for a in origin_aliquots:
                oas[a.id] = a
            self.failUnless(12345 in oas and 23456 in oas)
            self.failUnless(oas[12345].weight == 1.0)
            self.failUnless(oas[23456].container_id == self.container.id)
            aug_sample = IAugmentedSample(sample)
            self.failUnless(aug_sample.hol == u'4')
            sample = ISampleWithAliquots(sample)
            aliquots = sample.aliquots.values()
            self.failUnless(len(aliquots) == 1)
            self.failUnless(aliquots[0].weight == 0.5)
            self.failUnless(aliquots[0].container_id == self.container.id)
        except:
            print 'log:\n' + log
            raise




order = (
    Test_AddSampleAliquot,
    Test_ModifyAliquot,
    TestRoundtripModifyAliquot,
    Test_SplitAliquot,
    Test_DeriveSample,
    Test_AddContainer,
    Test_ModifyContainer,
    )

def test_suite():
    suite = unittest.TestSuite()
    for klass in order:
        f_name = '%s%s' % (klass.prefix, klass.fname)
        data = file(join(_prefix, 'data', f_name), 'r')
        klass.data = data.read()
        data.close()
        bad_data = file(join(_prefix, 'bad_data', f_name), 'r')
        klass.bad_data = bad_data.read()
        bad_data.close()
        stripped = file(join(_prefix, 'stripped', f_name), 'r')
        klass.stripped = stripped.read()
        stripped.close()
        suite.addTest(unittest.makeSuite(klass))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
