# Copyright (c) 2004 by RUNYAGA, LLC.
# http://www.runyaga.com/

"""
$Id: utils.py 749 2005-06-28 22:21:49Z sidnei $
"""

import csv
import glob
from os.path import split, join
from types import StringTypes
from cStringIO import StringIO
from pprint import pformat
from copy import deepcopy
from tempfile import TemporaryFile

from zope.app import zapi
from zope.schema import getFieldNames, getFieldsInOrder
from zope.security.proxy import removeSecurityProxy

from sip.browser import api
from sip.interfaces.adapter import IAugmentedSample, \
     ISampleWithAliquots, IZopeAliquot

class Table(object):

    def __init__(self, name, keys):
        self._name = name
        self._keys = keys
        self._rows = []
        self._count = 0

    def name(self):
        return self._name

    def __iter__(self):
        """Hail to iterators"""
        for r in iter(self._rows):
            # Don't you love python for allowing this?
            yield dict(zip(self._keys, r))

    def keys(self):
        """Table keys, a.k.a. column names"""
        return self._keys

    def values(self):
        """Only rows"""
        return self._rows

    def items(self):
        """For the lazy"""
        return list(self)

    def rename(self, name, new_name):
        """Rename a column"""
        keys = self._keys
        if not name in keys:
            raise KeyError, name
        keys[keys.index(name)] = new_name

    def copy(self, empty=False):
        new = Table(self._name, deepcopy(self._keys))
        if not empty:
            new.extend(deepcopy(self.values()))
        return new

    def extend(self, rows):
        """Pretend I am a list"""
        self._rows.extend(rows)
        self._count += len(rows)

    def index(self, name):
        return self._keys.index(name)

    def append(self, row):
        """Pretend I am a list"""
        self._rows.append(row)
        # Keep our own count, so we don't have
        # to len(self._rows) everytime
        self._count += 1
        return self._count - 1

    def insert(self, position, row):
        """Pretend I am a list"""
        self._rows.insert(position, row)
        # Keep our own count, so we don't have
        # to len(self._rows) everytime
        self._count += 1
        return self._count - 1

    def update(self, index, name, value):
        """Update one column, by index"""
        if name not in self._keys:
            raise KeyError, name
        ci = self._keys.index(name)
        self._rows[index][ci] = value

    def delete(self, index):
        """Delete by index"""
        del self._rows[index]
        self._count -= 1

    def select(self, i=True, single=True, **kw):
        """Select row where name == value.

        If i is False, then do a case-sensitive comparison"""
        for k in kw:
            if k not in self._keys:
                raise KeyError, k
        k = kw.keys()[0]
        index = self._keys.index(k)
        if i: value = str(kw[k]).lower()
        results = []
        for i, r in enumerate(self._rows):
            v = r[index]
            if str(v).lower() == value:
                results.append((i+1, self[i]))
        return results and single and results[0] or results or None

    def __len__(self):
        """Number of rows"""
        return self._count

    def __getitem__(self, index):
        r = self._rows[index]
        return dict(zip(self._keys, r))

    def __delitem__(self, name):
        """Nifty, delete columns"""
        if name not in self._keys:
            raise KeyError, name
        index = self._keys.index(name)
        for r in iter(self._rows):
            del r[index]
        del self._keys[index]

    def __setitem__(self, name, value):
        """Create column, with default value"""
        if name in self._keys:
            raise KeyError, name
        self._keys.append(name)
        for i, r in enumerate(self._rows):
            v = value
            if type(v) in StringTypes:
                v = value % {'count': i,
                             'id': i+1}
            r.append(v)

    def __repr__(self):
        """Very informative"""
        return ('<%s %s>' %
                (self.__class__.__name__,
                 pformat({'name':self._name,
                          'keys':self._keys,
                          'rows':len(self)}))
                )

class Exc(object):

    def __init__(self):
        self._exc = {}

    def get(self):
        return self._exc


    def keys(self):
        return self._exc.keys()

    def values(self):
        return self._exc.values()

    def items(self):
        return self._exc.items()

    def printOne(self, table, index=0):
        print self._exc[table][index][1]

    def printNumbers(self):
        print "Exceptions:"
        for k, v in self._exc.items():
            print k, len(v)

    def __setitem__(self, k, v):
        self._exc[k] = v

    def __getitem__(self, k):
        return self._exc[k]

class Chunk(object):

    def __init__(self):
        self.exceptions = Exc()
        self.tables = {}

    def __setitem__(self, k, v):
        self.tables[k] = v

    def __getitem__(self, k):
        return self.tables[k]

    def keys(self):
        return self.tables.keys()

    def values(self):
        return self.tables.values()

    def items(self):
        return self.tables.items()

    def get(self, k, default):
        return self.tables.get(k, default)

    def __str__(self):
        out = StringIO()
        print >> out, '<%s' % self.__class__.__name__
        print >> out, ' tables: %s' % pformat(self.values())
        print >> out, ' exceptions: %s>' % pformat(self.exceptions.get()),
        return out.getvalue()

def csv_read_dir(prefix):
    input_files = glob.glob(join(prefix, "*.csv"))
    tables = {}
    for f in input_files:
        l = None
        table = None
        keys = None
        warnings = []
        name = split(f)[-1].split('.')[0]
        infile = open(f)
        read_csv(infile, name, tables, warnings)
        if warnings:
            lines = '\n'.join([str(r) for r in warnings])
            print ("%s\nWarning: the preceding %s lines from %s "
                   "have a different number of fields as the rest "
                   "of the file." % (lines, len(warnings), name))
    return tables

def read_csv(infile, name, tables, warnings):
    reader = csv.reader(infile, skipinitialspace=True)
    table = None
    l = None
    for row in reader:
        if table is None:
            # Make sure there are no columns with
            # an empty name. This may mean that an
            # external app added trailing commas to
            # pad the number of columns across the file.
            row = [r.strip() for r in row]
            row = filter(None, row)
            tables[name] = table = Table(name, row)
            l = len(row)
            continue
        row = [r.strip() for r in row]
        if len(row) < l:
            for i in range(l - len(row)):
                row.append('')
            warnings.append(row)
        elif len(row) > l:
            row = row[:l]
            warnings.append(row)
        # Make sure at least one column in
        # the row has a value before append
        # to table. This can be silently ignored.
        if filter(None, row):
            table.append(row)
    return warnings


class AdditionalDataGetter:

    def __init__(self, context):
        self._cache = {}
        self.fieldFactory = api.factory('sip.SchemaField', context=context)
        self.schemaFactory = api.factory('sip.SampleTypeSchema',
                context=context)
        self.dataFactory = api.factory('sip.AdditionalSampleData',
                context=context)

    def getDataForSample(self, sample):
        cache = self._cache
        type_id = sample.type_id
        project_id = sample.project_id
        project_name = sample.project_.name
        sample_name = sample.name
        sample_id = sample.id

        fields = cache.get((type_id, project_id), None)
        if fields is None:
            fields = {}
            factory = self.schemaFactory
            query = 'project_id = %d and sample_type_id = %d' % (project_id,
                    type_id)
            results = factory.select(clause=query)
            assert results.count() == 1
            schema_id = results[0].id
            factory = self.fieldFactory
            query = 'schema_id = %d' % schema_id
            sqlfields = factory.select(clause=query)
            for field in sqlfields:
                f = fields[field.id] = {}
                f['title_or_id'] = field.title or field.name
                f['default'] = field.default
            cache[(type_id, project_id)] = fields
        if not fields: return []
        factory = self.dataFactory
        field_ids = fields.keys()
        if len(field_ids) > 1:
            query = 'sample_id = %d and schema_field_id in %r' % (sample_id,
                    tuple(field_ids))
        else:
            query = 'sample_id = %d and schema_field_id = %d' % (sample_id,
                    field_ids[0])
        results = factory.select(clause=query)
        data = []
        data_dict = {}
        for res in results:
            data_dict[res.schema_field_id] = res.value
        for field_id, field in fields.items():
            d = [project_name, sample_id, sample_name]
            d.append(field['title_or_id'])
            d.append(data_dict.get(field_id, field['default'] or ''))
            data.append(d)
        return data

def dump_samples(samples, as_string=True):
    out = {}
    out['sample'] = sample_table = Table('sample',
                                         ['project_name',
                                          'sample_id',
                                          'sample_name',
                                          'sample_type',
                                          'subject_id',
                                          'subject_type',
                                          'other_id',
                                          'technician',
                                          'collection_date',
                                          'comments'])

    out['additional_data'] = ad_table = Table('additional_data',
                                              ['project_name',
                                               'sample_id',
                                               'sample_name',
                                               'field_title_or_id',
                                               'field_value'])

    out['aliquots'] = aliquot_table = Table('aliquots',
                                            ['project_name',
                                             'aliquot_id',
                                             'sample_id',
                                             'sample_name',
                                             'weight',
                                             'weight_units',
                                             'volume',
                                             'volume_units',
                                             'concentration',
                                             'concentration_units',
                                             'chiclet_count',
                                             'is_low',
                                             'depleted',
                                             'container_path',
                                             'comment'])
    project_cache = {}
    schema_cache = {}
    data_getter = AdditionalDataGetter(api.getProject(samples[0]))
    for sample in samples:
        sample = removeSecurityProxy(sample)
        sdata = []
        project_name = sample.project_.name
        sample_name = sample.name
        sample_id = sample.id
        add = sdata.append
        add(project_name)
        add(sample_id)
        add(sample_name)
        add(sample.type_.name)
        add(sample.subject_id)
        add(sample.subject_type_.name)
        add(sample.other_id)
        tech = sample.technician_
        add(tech and tech.name or '')
        add(sample.collection_date)
        add(sample.comments)
        sample_table.append(sdata)

        ad_data = data_getter.getDataForSample(sample)

        for row in ad_data:
            ad_table.append(row)

        # No need to use the output of dump_aliquots
        # because we are passing aliquot_table, which
        # will be appended to inside the function.
        dump_aliquots(sample.aliquots.values(),
                      aliquot_table=aliquot_table,
                      as_string=False)

    if as_string:
        return tables_as_string(out)
    return out

def dump_aliquots(aliquots, aliquot_table=None,
                  as_string=True):
    out = {}

    if aliquot_table is not None:
        out['aliquots'] = aliquot_table
    else:
        out['aliquots'] = aliquot_table = Table('aliquots',
                                                ['project_name',
                                                 'aliquot_id',
                                                 'sample_id',
                                                 'sample_name',
                                                 'weight',
                                                 'weight_units',
                                                 'volume',
                                                 'volume_units',
                                                 'concentration',
                                                 'concentration_units',
                                                 'chiclet_count',
                                                 'is_low',
                                                 'depleted',
                                                 'container_path',
                                                 'comment'])
    path_cache = {}
    cunit_cache = {}
    vunit_cache = {}
    wunit_cache = {}
    sample_cache = {}
    project_cache = {}
    for aliquot in aliquots:
        al_data = []
        add = al_data.append
        project_id = aliquot.project_id
        project_name = project_cache.get(project_id, None)
        if project_name is None:
            project = aliquot.project_
            project_name = project_cache[project_id] = project.name
        add(project_name)
        add(aliquot.id)
        sample_id = aliquot.sample_id
        sample_name = sample_cache.get(sample_id, None)
        if sample_name is None:
            sample_name = sample_cache[sample_id] = aliquot.sample_.name
        add(sample_id)
        add(sample_name)
        add(aliquot.weight)
        wunit_id = aliquot.weight_id
        w = wunit_cache.get(wunit_id, None)
        if w is None:
            if wunit_id:
                w = aliquot.weight_
                w = wunit_cache[wunit_id] = w.name
            else:
                w = wunit_cache[wunit_id] = ''
        add(w)
        add(aliquot.volume)
        vunit_id = aliquot.volume_id
        v = vunit_cache.get(vunit_id, None)
        if v is None:
            if vunit_id:
                v = aliquot.volume_
                v = vunit_cache[vunit_id] = v.name
            else:
                v = vunit_cache[vunit_id] = ''
        add(v)
        add(aliquot.concentration)
        cunit_id = aliquot.concentration_id
        c = cunit_cache.get(cunit_id, None)
        if c is None:
            if cunit_id:
                c = aliquot.concentration_
                c = cunit_cache[cunit_id] = c.name
            else:
                c = cunit_cache[cunit_id] = ''
        add(c)
        add(aliquot.aliquot_count)
        add(aliquot.is_low)
        add(aliquot.depleted)
        container_id = aliquot.container_id
        path = None
        if container_id: path = path_cache.get(container_id, None)
        if path is None:
            path = aliquot.getLocation()
            if path is not None:
                names = [d['title'].strip() for d in path]
                path = "path://%s" % '/'.join(names)
            path = path_cache[container_id] = path or ''

        add(path)
        add(aliquot.comment or '')
        aliquot_table.append(al_data)

    if as_string:
        return tables_as_string(out)
    return out

def tables_as_string(out):
    for name, table in out.items():
        tmp = TemporaryFile()
        writer = csv.writer(tmp)
        writer.writerow(table.keys())
        writer.writerows(table.values())
        tmp.seek(0)
        out[name] = tmp.read()
        tmp.close()
    return out
