# Copyright (c) 2004 by RUNYAGA, LLC.
# http://www.runyaga.com/

"""
$Id: base.py 759 2005-06-28 23:45:42Z sidnei $
"""

from transaction import get_transaction

from zope.publisher.browser import TestRequest
from zope.publisher.base import DefaultPublication
from zope.component.exceptions import ComponentLookupError
from zope.security.management import system_user, newInteraction, endInteraction
from zope.app.tests.functional import FunctionalTestCase

from zope.app import zapi
from zope.app.tests import setup
from zope.app.servicenames import Authentication
from zope.app.container.interfaces import IReadContainer

from sip.root import AppRoot
from sip.browser import api


class ParticipationStub:

    def __init__(self, principal):
        self.principal = principal
        self.interaction = None


class TestFailure(Exception):
    pass

class Base(FunctionalTestCase):
    # ensure a failed "assert" in the code is signaled as an
    # error, not a failure
    failureException = TestFailure

    # attributes used for test class setup
    full_setup = False
    use_test_user = True
    create_approot = True

    def setUp(self):
        super(Base, self).setUp()
        self._setupManagerUser()
        self.rootFolder = setup.buildSampleFolderTree()

        if self.create_approot:
            # Add approot
            app = AppRoot()
            self.rootFolder['sip'] = self.sip = app
            # Save tables
            get_transaction().commit()


        # Create a publication for testing
        publication = DefaultPublication(self.rootFolder)
        # And a request
        self.request = TestRequest(skin='sip')
        self.request.setPublication(publication)
        factory = api.factory('sip.SubjectType', context=app)
        self.animal = factory.get(1)
        self.other = factory.get(2)

        try:
            if self.full_setup:
                self._fullSetup()
            if self.use_test_user:
                # Then switch to the test user
                self._setupTestUser()
        except:
            # ... and abort in case we get an exception during setup.
            get_transaction().abort()
            raise

    def _setupSystemUser(self):
        endInteraction()
        newInteraction(ParticipationStub(system_user))

    def _setupManagerUser(self):
        auth_service = zapi.getService(Authentication)
        user = auth_service.getPrincipal('zope.mgr')
        endInteraction()
        newInteraction(ParticipationStub(user))

    def _setupTestUser(self):
        endInteraction()
        newInteraction(ParticipationStub(system_user))

    def _fullSetup(self):
        project = self.project = self._addProject()
        ct = self.container_type = self._addContainerType()
        locations = zapi.traverse(self.sip, 'locations')
        location = self.location = self._addLocation(type_id=ct.id)
        self._addContainer(location, ct.id)
        technician = self.technician = self._addTechnician()
        cunit = self.cunit = self._addConcentrationUnit()
        vunit = self.vunit = self._addVolumeUnit()
        wunit = self.wunit = self._addWeightUnit()
        locations = zapi.traverse(project, 'locations')
        project.valid_locations = [location.id]
        project.valid_technicians = [technician.id]
        aliquots = zapi.traverse(project,'aliquots')
        stype = self.sample_type = self._addSampleType(project)
        sample = self.sample = self._addSample(project, stype.id,
                                               technician.id)
        aliquot = self.aliquot = self._addAliquot(project, sample.id,
                volume = 1.0, weight = 2.0,
                volume_id = vunit.id, weight_id = wunit.id)
        self.ids = {}
        self.ids['project'] = project.id
        self.ids['sampletype'] = stype.id
        self.ids['sample'] = sample.id
        self.ids['aliquot'] = aliquot.id
        self.ids['subjecttype'] = self.animal.id
        self.ids['location'] = location.id
        self.ids['technician'] = technician.id
        self.ids['cunit'] = cunit.id
        self.ids['vunit'] = vunit.id
        self.ids['wunit'] = wunit.id
        # Add another project for 'SeeProjects' test.
        project2 = self._addProject(name=u'Proyecto Dos')


    def _addContainerType(self, **kwargs):
        containertypes = zapi.traverse(self.sip, 'containertypes')
        adding = zapi.getView(containertypes, '+', self.request)
        adding = zapi.getView(adding, 'add_container_type.html', self.request)

        kw = {}
        kw['name'] = u'Freezer'
        kw['root'] = True
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        ct = adding.create(**kw)
        return containertypes['sip.ContainerType.%s' %  ct.id]

    def _addLocation(self, type_id=None, **kwargs):
        locations = zapi.traverse(self.sip, 'locations')
        adding = zapi.getView(locations, '+', self.request)
        adding = zapi.getView(adding, 'add_location.html', self.request)

        kw = {}
        kw['name'] = u'FreezerA'
        kw['type_id'] = type_id
        kw['building'] = u'Building #10'
        kw['room'] = u'Room #2'
        kw['temperature'] = 10
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        loc = adding.create(**kw)
        return locations['sip.Location.%s' % loc.id]

    def _addTechnician(self, **kwargs):
        technicians = zapi.traverse(self.sip, 'technicians')
        adding = zapi.getView(technicians, '+', self.request)
        adding = zapi.getView(adding, 'add_technician.html', self.request)

        kw = {}
        kw['username'] = u'alan'
        kw['fullname'] = u'George Alan Runyan'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        tech = adding.create(**kw)
        return technicians['sip.Technician.%s' % tech.id]

    def _addContainer(self, location, type_id, **kwargs):
        location = IReadContainer(location)
        adding = zapi.getView(location, '+', self.request)
        adding = zapi.getView(adding, 'add_container.html', self.request)

        kw = {}
        kw['name'] = 'box'
        kw['rows'] = 1
        kw['columns'] = 1
        kw['type_id'] = type_id
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        ct = adding.create(**kw)
        location = IReadContainer(location)
        return location['sip.Container.%s' % ct.id]

    def _addProject(self, **kwargs):
        projects = zapi.traverse(self.sip, 'projects')
        adding = zapi.getView(projects, '+', self.request)
        adding = zapi.getView(adding, 'add_project.html', self.request)

        kw = {}
        kw['name'] = u'Proyecto'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        proj = adding.create(**kw)
        return projects['sip.Project.%s' % proj.id]

    def _addConcentrationUnit(self, **kwargs):
        request = TestRequest()
        cunits = zapi.traverse(self.sip, 'concentrationunits')
        adding = zapi.getView(cunits, '+', self.request)
        adding = zapi.getView(adding, 'add_concentration_unit.html',
                              self.request)

        kw = {}
        kw['exponent'] = 1
        kw['label'] = 'testCU'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        return adding.create(**kw)

    def _addVolumeUnit(self, **kwargs):
        request = TestRequest()
        vunits = zapi.traverse(self.sip, 'volumeunits')
        adding = zapi.getView(vunits, '+', self.request)
        adding = zapi.getView(adding, 'add_volume_unit.html', self.request)

        kw = {}
        kw['exponent'] = 1
        kw['label'] = 'testVU'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        return adding.create(**kw)

    def _addWeightUnit(self, **kwargs):
        request = TestRequest()
        wunits = zapi.traverse(self.sip, 'weightunits')
        adding = zapi.getView(wunits, '+', self.request)
        adding = zapi.getView(adding, 'add_weight_unit.html', self.request)

        kw = {}
        kw['exponent'] = 1
        kw['label'] = 'testWU'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        return adding.create(**kw)

    def _addSampleType(self, project, **kwargs):
        stypes = zapi.traverse(self.sip, 'sampletypes')
        adding = zapi.getView(stypes, '+', self.request)
        adding = zapi.getView(adding, 'add_sample_type.html', self.request)

        kw = {}
        kw['name'] = 'blood'
        kw['commit_message'] = u'test'
        if kwargs:
            kw.update(kwargs)

        st = adding.create(**kw)
        valid = list(project.valid_sample_types)
        valid.append(st.id)
        project.valid_sample_types = valid
        return stypes['sip.SampleType.%s' % st.id]

    def _addSample(self, project, stype_id, tech_id=None, **kwargs):
        samples = zapi.traverse(project, 'samples')
        adding = zapi.getView(samples, '+', self.request)
        adding = zapi.getView(adding, 'add_sample.html', self.request)

        kw = {}
        kw['name'] = 'testSample'
        kw['subject_id'] = u'1'
        kw['other_id'] = u''
        kw['technician_id'] = tech_id
        kw['comments'] = u''
        kw['type_id'] = stype_id
        kw['subject_type_id'] = self.animal.id
        kw['commit_message'] = u'test'
        kw['collection_date'] = u'1/1/1970'
        if kwargs:
            kw.update(kwargs)

        sample = adding.create(**kw)
        return samples['sip.Sample.%s' % sample.id]

    def _addAliquot(self, project, sample_id, **kwargs):
        aliquots = zapi.traverse(project,'aliquots')

        kw = {}
        kw['id'] = api.allocateAliquotIds(project)[0]
        kw['volume_id'] = None
        kw['concentration_id'] = None
        kw['weight_id'] = None
        kw['container_id'] = None
        kw['volume'] = None
        kw['concentration'] = None
        kw['weight'] = None
        kw['sample_id'] = sample_id
        kw['project_id'] = project.id
        kw['commit_message'] = u'test'
        kw['comment'] = None

        adding = zapi.getView(aliquots, '+', self.request)
        adding = zapi.getView(adding, 'add_aliquot.html', self.request)

        if kwargs:
            kw.update(kwargs)

        al = adding.create(**kw)
        return aliquots['sip.Aliquot.%d' % al.id]

    def _addPrincipal(self, **kwargs):
        principals = zapi.traverse(self.sip, 'principals')
        adding = zapi.getView(principals, '+', self.request)
        adding = zapi.getView(adding, 'add_principal.html', self.request)

        kw = {}
        kw['username'] = u'test_user'
        kw['fullname'] = u'Test User'
        kw['password'] = u'testpass'
        kw['confirm_pass'] = u'testpass'
        if kwargs:
            kw.update(kwargs)

        princ = adding.create(**kw)
        return principals['sip.Principal.%s' % princ.id]

    def _addShipment(self, context, **kwargs):
        shipments = zapi.traverse(context, 'shipments/outgoing/incomplete')
        adding = zapi.getView(shipments, '+', self.request)
        adding = zapi.getView(adding, 'add_shipment.html', self.request)
        kw = {'planneddate':None,
              'requestdate':None,
              'shipdate':None,
              'receivedate':None,
              'returndate':None,
              'biohazard':False,
              'infectious':False,
              'commit_message': 'test',
              }
        if kwargs:
            kw.update(kwargs)
        ship = adding.create(**kw)
        return shipments['sip.Shipment.%s' % ship.id]



    def tearDown(self):
        super(Base, self).tearDown()
        get_transaction().abort()
