# Copyright (c) 2004 by RUNYAGA, LLC.
# http://www.runyaga.com/

"""
$Id: test_sample.py 749 2005-06-28 22:21:49Z sidnei $
"""

import unittest
from datetime import date
from sqlobject import *

from zope.app import zapi
from zope.schema import Text, Bool, Date, Int, Float
from zope.security.proxy import ProxyFactory
from zope.app.datetimeutils import parse
from sip.browser import api
from sip.ftests.base import Base
from sip.interfaces.adapter import IAugmentedSample, \
     ISampleWithAliquots, ISampleWithOriginAliquots
from sip.interfaces.fields import ITALESConstrainedField

class TestSample(Base):
    level = 2
    use_test_user = False

    def _addSampleType_with_Schema(self):
        project = self.project
        stypes = zapi.traverse(project, 'sampletypes')
        self.assertEquals(len(stypes), 0)
        stype = self._addSampleType(project)
        self.failUnless(stype.id == iter(stypes.values()).next().id)
        schema = api.getSchemaFor(stype.id, project)
        mofo = Text(title=u'Mofo', required=True, default=u'foo')
        schema.addField(u'mofo', mofo)
        schema.constrain(u'mofo', u'python:value != u"invalid"')
        mofo = schema.get('mofo')
        self.failUnless(ITALESConstrainedField.providedBy(mofo))
        alcoh = Bool(title=u'Alcoholic', required=True, default=True)
        schema.addField(u'alcoholic', alcoh)
        bday = Date(title=u'Birthday', required=False)
        schema.addField(u'bday', bday)
        age = Int(title=u'Age', required=False)
        schema.addField(u'age', age)
        salary = Float(title=u'Salary', required=False)
        schema.addField(u'salary', salary)
        return stype

    def testSampleType(self):
        project = self._addProject()
        stypes = zapi.traverse(project, 'sampletypes')
        self.assertEquals(len(stypes),0)
        stype = self._addSampleType(project)
        self.failUnless(stype.id == iter(stypes.values()).next().id)

    def testSample(self):
        project = self._addProject()
        samples = zapi.traverse(project, 'samples')
        self.assertEquals(len(samples),0)
        tech = self._addTechnician()
        stype = self._addSampleType(project)
        sample = self._addSample(project, stype.id,
                                 tech.id, name='')
        nsample = iter(samples.values()).next()
        self.failUnless(sample.id == nsample.id)
        self.failUnless(sample.name == str(sample.id))
        sample.name = 'testSample'
        self.failUnless(sample.name == 'testSample', sample.name)
        sample.name = ''
        self.failUnless(sample.name == str(sample.id))

    def testDeriveSample(self):
        project = self._addProject()
        tech = self._addTechnician()
        stype = self._addSampleType(project)
        sample = self._addSample(project, stype.id, tech.id)
        aliquot = self._addAliquot(project, sample.id)
        aliquot2 = self._addAliquot(project, sample.id)
        alids = [aliquot.id, aliquot2.id]
        samples = zapi.traverse(project, 'samples')
        adding = zapi.getView(samples, '+', self.request)
        self.request.form['aliquot_ids'] = alids
        adding = zapi.getView(adding, 'DerivedSample', self.request)
        sip = zapi.traverse(self.rootFolder, 'sip')
        osa = zapi.traverse(sip, 'originsamplealiquots')
        samples = zapi.traverse(project, 'samples')
        self.failUnless(len(samples) == 1)
        kw = {}
        kw['name'] = u'derived'
        kw['subject_id'] = 1
        kw['other_id'] = u''
        kw['technician_id'] = tech.id
        kw['comments'] = u''
        kw['type_id'] = stype.id
        kw['subject_type_id'] = self.animal.id
        kw['commit_message'] = u'test'
        kw['collection_date'] = date.today()
        newsample = adding.create(**kw)
        newsample = api.contained(newsample, samples,
                                  'sip.Sample.%d' % newsample.id)
        adapted = ISampleWithOriginAliquots(newsample)
        origin_aliquots = adapted.origin_aliquots.values()
        self.failUnless(len(origin_aliquots) == 2)
        self.failUnless(len(samples) == 2)
        self.failUnless(len(osa) == 2)
        origins = [v for v in osa.values()]
        self.failUnless(origins[0].sample_id == newsample.id)
        self.failUnless(origins[1].sample_id == newsample.id)
        self.failUnless(origins[0].aliquot_id != origins[1].aliquot_id)
        self.failUnless(origins[0].aliquot_id in alids)
        self.failUnless(origins[1].aliquot_id in alids)
        view = zapi.getView(newsample, 'edit_sample.html', self.request)
        self.failUnless(view())

    def testEditSampleWithFields(self):
        self.project = project = self._addProject()
        tech = self._addTechnician()
        stype = self._addSampleType_with_Schema()
        sample = self._addSample(project, stype.id, tech.id)
        form = self.request.form
        form['field.mofo'] = u'nay'
        form['field.alcoholic.used'] = True
        form['field.commit_message'] = 'Test'
        form['field.bday'] = u'10/25/2004'
        form['field.age'] = u'29'
        form['field.salary'] = u'30000'
        form['UPDATE_SUBMIT'] = 'UPDATE_SUBMIT'
        url = str(zapi.getView(sample, 'absolute_url', self.request))
        self.request._traversed_names = url.split('/')[3:]
        editsample = zapi.getView(sample, 'edit.html', self.request)
        editsample.update()
        self.failIf(editsample.errors)
        aug = IAugmentedSample(sample)
        self.failUnless(aug.mofo == u'nay')
        self.failUnless(aug.alcoholic == False, repr(aug.alcoholic))
        test_date = parse(u'10/25/2004')
        test_date = date(test_date[0], test_date[1], test_date[2])
        self.failUnless(aug.bday == test_date)
        self.failUnless(aug.age == int(u'29'))
        self.failUnless(aug.salary == float(u'30000'))

    def testEditSampleWithMissingFields(self):
        self.project = project = self._addProject()
        tech = self._addTechnician()
        stype = self._addSampleType_with_Schema()
        sample = self._addSample(project, stype.id, tech.id)
        form = self.request.form
        form['field.mofo'] = u''
        form['field.commit_message'] = 'Test'
        form['UPDATE_SUBMIT'] = 'UPDATE_SUBMIT'
        url = str(zapi.getView(sample, 'absolute_url', self.request))
        self.request._traversed_names = url.split('/')[3:]
        editsample = zapi.getView(sample, 'edit.html', self.request)
        editsample.update()
        self.failUnless(editsample.errors)

    def testEditSampleWithConstraint(self):
        self.project = project = self._addProject()
        tech = self._addTechnician()
        stype = self._addSampleType_with_Schema()
        sample = self._addSample(project, stype.id, tech.id)
        form = self.request.form
        form['field.mofo'] = u'invalid'
        form['field.alcoholic.used'] = True
        form['field.commit_message'] = 'Test'
        form['field.bday'] = u'10-25-2004'
        form['field.age'] = u'29'
        form['field.salary'] = u'30000'
        form['UPDATE_SUBMIT'] = 'UPDATE_SUBMIT'
        url = str(zapi.getView(sample, 'absolute_url', self.request))
        self.request._traversed_names = url.split('/')[3:]
        editsample = zapi.getView(sample, 'edit.html', self.request)
        editsample.update()
        self.failUnless(editsample.errors)

    def testAdaptSampleToSampleWithAliquots(self):
        self.project = project = self._addProject()
        tech = self._addTechnician()
        stype = self._addSampleType(project)
        sample = self._addSample(project, stype.id, tech.id)
        aliquot = self._addAliquot(project, sample.id)
        adapted = ISampleWithAliquots(sample)
        self.failUnless(adapted.aliquots)
        self.failUnless(len(adapted.aliquots) == 1)
        self.failUnless(adapted.aliquots.values()[0].id == aliquot.id)

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestSample))
    return suite

if __name__ == '__main__':
    unittest.main()
