##############################################################################
#
# Copyright (c) 2004 Enfold Systems LLC. All rights reserved.
#
# This software is distributed under the terms of the Zope Public
# License (ZPL) v2.1. See COPYING.txt for more information.
#
##############################################################################
"""
$Id: adapter.py 5212 2004-06-21 18:09:05Z philikon $
"""
from zope.schema import TextLine
from zope.interface import Interface, Attribute
from sqlobject import NoDefault
from zope.component import getService, ComponentLookupError
from zope.schema.vocabulary import SimpleVocabulary
from zope.schema import Choice, List
from zope.app.annotation.interfaces import IAttributeAnnotatable

class IConnectionName(Interface):
    """A marker interface for providing a connection name"""

    name = TextLine(
        title=u"Connection Name",
        required=True
        )

class ISQLSchema(Interface):
    """ SQLObject-based schemas must declare 'id' or else it doesn't
    get security set """

    id = Attribute('Id')

class IDBConnection(Interface):
    """ SQLObject DBConnection interface """

    name  = Attribute("The object name, used as key for caching connections")
    debug = Attribute("Print debug trace messages")
    cache = Attribute("A ResultSet cache object")
    style = Attribute("A style object. Used for controlling the naming style.")

class IDBAPI(IDBConnection):

    def _runWithConnection(meth, *args):
        """ Runs a method with a connection from the pool """

    def getConnection():
        """ Return a connection from the pool """

    def releaseConnection(conn):
        """ Return a connection back to the pool """

    def query(s):
        """ Run a query string against the database """

    def queryAll(s):
        """ Run a query string against the database and return all results """

    def queryOne(s):
        """ Run a query string against the database and return one row """

    def transaction():
        """ Return a transaction object for this connection """

    def queryInsertID(table, idName, id, names, values):
        """ Insert a row into the database and return the generated id """

    def iterSelect(select):
        """ """

    def queryForSelect(select):
        """ """

    def createTable(soClass):
        """ Create a table for the given SQLObject class """

    def createColumns(soClass):
        """ Return a query needed to create the columns for the given SO class """

    def dropTable(tableName):
        """ Drop the given table """

    def clearTable(tableName):
        """ Clear the given table """

    ### Private methods

    def _SO_update(so, values):
        """ """

    def _SO_selectOne(so, columnNames):
        """ """

    def _SO_selectOneAlt(cls, columnNames, column, value):
        """ """

    def _SO_delete(so):
        """ """

    def _SO_selectJoin(soClass, column, value):
        """ """

    def _SO_intermediateJoin(table, getColumn, joinColumn, value):
        """ """

    def _SO_intermediateDelete(table, firstColumn, firstValue,
                               secondColumn, secondValue):
        """ """

    def _SO_intermediateInsert(table, firstColumn, firstValue,
                               secondColumn, secondValue):
        """ """

    def _SO_columnClause(soClass, kw):
        """ """

class ISQLConnection(IDBAPI):

    def makeConnection():
        """ Return a newly-built connection instance using args passed
        on __init__ """

    def getConnection():
        """ Get a connection from the pool """

    def _runWithConnection(meth, *args):
        """ Run a method with the give args and a connection """

    def createColumn(soClass, column):
        """ Create a column """

    def createIDColumn(soClass):
        """ Return the string used to create the ID column """

    def joinSQLType(join):
        """ Return the string to be used in join queries """

    def tableExists(tableName):
        """ Return if the given table exists or not in the database """

    def addColumn(tableName, column):
        """ Add the given column to the database """

    def delColumn(tableName, column):
        """ Remove the given column from the database """

    def columnsFromSchema(tableName, soClass):
        """ Return a set of columns from the table schema """

    def guessClass(t):
        """ Returns a column class and a dict to be used on column
        initialization """

class IZopeSQLConnection(ISQLConnection):
    """ """

class ISQLAttributeAnnotatable(IAttributeAnnotatable):
    """
    Store annotations in the annotations table, keyed by
    table_name/id on a IAttributeAnnotatable object.
    """
class IReadSQLObjectClass(Interface):

    q = Attribute('Query?')

    def get(id):
        """Return object by the given primary key.
        """

    def sqlrepr(value):
        """ Shorthand for _connection.sqlrepr
        """

    def select(clause=None, clauseTables=None, orderBy=NoDefault,
               groupBy=None, limit=None, lazyColumns=False,
               reversed=False):
        """ Do a select query and return the resulting objects.
        """

    def selectBy(**kw):
        """ """

class IWriteSQLObjectClass(Interface):


    def delete(id):
        """ Delete item by primary key id.
        """

    def addColumn(columnDef, changeSchema=False):
        """ Add a column to the class. If changeSchema is True, also
        add the column to the database. Also generates getter and
        setter methods on the class.
        """

    def addColumnsFromDatabase():
        """ Add to the class the columns that are defined on the
        database but not already present.
        """

    def delColumn(column, changeSchema=False):
        """ Delete the given column from the class. 'column' may be
        either a string or a Col instance. If changeSchema is true,
        also remove the column from the database table.
        """

    def addJoin(joinDef):
        """ Add a join definition to the class, optionally removing or
        adding items as requested.
        """

    def delJoin(joinDef):
        """ Remove a join definition from the class, optionally
        removing or adding items as requested.
        """

    def dropTable(ifExists=False, dropJoinTables=True):
        """ Drop the table. If the 'ifExists' parameter was passed,
        check if the table exists before trying to delete. If
        'dropJoinTables' is True, drop the join tables associated.
        """

    def createTable(ifNotExists=False, createJoinTables=True):
        """ Create the table. If the 'ifNotExists' parameter was
        passed, check if the table exists before trying to create. If
        'createJoinTables' is True, create the join tables associated.
        """
    def createJoinTables(ifNotExists=False):
        """ Create the associated join tables. If the 'ifNotExists'
        parameter is True, check if the tables doesn't already exist
        first.
        """

    def dropJoinTables(ifExists=False):
        """ Drop the associated join tables. If the 'ifExists'
        parameter is True, then first check if the tables exist before
        trying to delete.
        """

    def clearTable():
        """ Clear the table.
        """

class IISQLObject(IReadSQLObjectClass, IWriteSQLObjectClass):
    """Class methods for SQLObject classes."""

class ISQLObject(Interface):

    _idName = Attribute('Primary Key')


    def set(**kw):
        """ Used to update multiple values at once, potentially with
        one SQL statement if possible.
        """

    def destroySelf():
        """ Kill this object and remove it from all caches. Also
        removes the row associated with this object from the table.
        """

    def syncUpdate():
        """ Submit changes to the database
        """

    def sync():
        """ If there are pending changes, submit them to the database,
        and after that re-sync the object with the database.
        """

    def expire():
        """ Expire the object, clearing the cache for the current connection.
        """

class ICacheSet(Interface):

    def get(id, cls):
        """ Get object 'id' from cache key 'cls.__name__' """

    def put(id, cls):
        """ Set a cache value using 'id' and 'cls.__name__' as keys """

    def finishPut(cls):
        """ """

    def created(id, cls, obj):
        """ """

    def purge(id, cls):
        """ Purge an object from the cache """

    def clear(cls=None):
        """ Clear the cache if cls is None, else clear only the given
        class cache """

class ISelectResults(Interface):

    def __getitem__(item):
        """ List Emulation """

    def __iter__():
        """ List Emulation """

    def __len__():
        """ List emulation """

class IIterator(Interface):

    def __iter__():
        """ Iterator """

    def next():
        """ Iterator """
