##############################################################################
#
# Copyright (c) 2004 Enfold Systems LLC. All rights reserved.
#
# This software is distributed under the terms of the Zope Public
# License (ZPL) v2.1. See COPYING.txt for more information.
#
##############################################################################
"""
$Id: __init__.py 12393 2005-05-16 22:06:40Z dreamcatcher $
"""
__metaclass__ = type

import sets

# The next import is needed so we don't shadow
# the global 'transaction' module from zope3.
from transaction import get_transaction

from transaction.interfaces import IDataManager, IRollback
from transaction.util import NoSavepointSupportRollback
from zope.interface import implements
from zope.app.event.objectevent import modified
from sqlos.interfaces import ISQLObject
from sqlos.event.interfaces import expired

class SQLObjectTransactionManager:
    """
    This is a very simple Data Manager that just takes registrations
    of ``ISQLObject`` objects and calls their ``sync()`` method when
    needed.

    In addition to that, when the transaction is aborted, all modified
    objects will be expired.

    Let's see how it works.

    First of all, setup the environment:

    >>> from zope.app.tests.placelesssetup import setUp, tearDown
    >>> setUp()

    First, register the subscribers:

    >>> from zope.app.tests import ztapi
    >>> from zope.app.event.interfaces import IObjectModifiedEvent
    >>> from zope.app.event.interfaces import IObjectCreatedEvent
    >>> from sqlos.event.interfaces import IObjectExpiredEvent
    >>> from sqlos.event.subscriber import sqlobjectModifiedSubscriber
    >>> from sqlos.event.subscriber import sqlobjectCreatedSubscriber
    >>> from sqlos.event.subscriber import sqlobjectExpiredSubscriber

    >>> ztapi.handle([IObjectExpiredEvent], sqlobjectExpiredSubscriber)
    >>> ztapi.handle([IObjectCreatedEvent], sqlobjectCreatedSubscriber)
    >>> ztapi.handle([IObjectModifiedEvent], sqlobjectModifiedSubscriber)

    Then, create a connection:

    >>> from sqlobject.dbconnection import connectionForURI
    >>> from sqlos.tests.sampleperson import SamplePerson
    >>> from sqlos.adapter import SQLiteAdapter
    >>> connection = connectionForURI('sqlite:///tmp/sqlite.db')
    >>> adapter = SQLiteAdapter(connection.makeConnection())

    Stuff the connection into the test class and initialize the table
    if it doesn't exists yet:

    >>> SamplePerson._connection = adapter
    >>> SamplePerson.createTable(ifNotExists=True)

    Now create a SamplePerson for testing:

    >>> person = SamplePerson(fullname='Sidnei', username='sidnei',
    ... password='123')
    >>> person.id
    1
    >>> person.fullname
    'Sidnei'
    >>> person.username
    'sidnei'
    >>> person.password
    '123'

    Create a temp method to ease testing:

    >>> oldsync = SamplePerson.syncUpdate
    >>> def syncUpdate(self):
    ...     self._sync_called = True
    ...     oldsync(self)
    ...
    >>> SamplePerson.syncUpdate = syncUpdate
    >>> SamplePerson._sync_called = False

    >>> oldexpire = SamplePerson.expire
    >>> def expire(self):
    ...     self._expire_called = True
    ...     oldexpire(self)
    ...
    >>> SamplePerson.expire = expire
    >>> SamplePerson._expire_called = False

    Now, we check the DataManager state:

    >>> person.dirty
    False
    >>> dm = SamplePerson._connection._dm
    >>> dm.state, dm.delta
    (0, 1)
    >>> person.set(fullname='Sidnei da Silva')
    >>> person.dirty
    True
    >>> dm.state, dm.delta
    (0, 2)

    Prepare the transaction, make sure we get the expected values:

    >>> t1 = '1'
    >>> dm.prepare(t1)
    True
    >>> dm.state, dm.delta
    (2, 2)

    Make sure prepare can't be called twice:

    >>> dm.prepare(t1)
    Traceback (most recent call last):
    ...
    TypeError: Already prepared

    Check that syncUpdate() hasn't been called yet:

    >>> person._sync_called
    False
    >>> person._expire_called
    False

    Commit, and make sure syncUpdate and expire have been called:

    >>> dm.commit(t1)
    >>> dm.state, dm.delta
    (0, 0)
    >>> person._sync_called
    True
    >>> person._expire_called
    True
    >>> person.dirty
    False

    Now, we do everything again to test abort():

    >>> person._sync_called = False
    >>> person._expire_called = False
    >>> person.set(fullname='Alan Runyan')
    >>> person.dirty
    True
    >>> dm.state, dm.delta
    (0, 1)
    >>> dm.prepare(t1)
    True
    >>> dm.state, dm.delta
    (1, 1)
    >>> dm.abort(t1)
    >>> dm.state, dm.delta
    (0, 0)

    Make sure syncUpdate and expire have been called:

    >>> person._sync_called
    True
    >>> person._expire_called
    True

    But the object is not dirty nor expired, because sync() has been
    called to refetch the values from the database:

    >>> person.dirty, person._expired
    (False, False)

    And then cleanup the monkeypatched method:

    >>> SamplePerson.syncUpdate = oldsync
    >>> del SamplePerson._sync_called
    >>> SamplePerson.expire = oldexpire
    >>> del SamplePerson._expire_called

    And finally call tearDown to cleanup the environment:

    >>> tearDown()
    """

    implements(IDataManager)

    def __init__(self, sqlconnection):
        self.objects = sets.Set()
        self.sqlconnection = sqlconnection
        self.state = 0
        self.prepared = False
        self.delta = 0
        self.transaction = None

    def _checkTransaction(self, txn):
        if (txn is not self.transaction
            and self.transaction is not None):
            raise TypeError("Transaction missmatch",
                            txn, self.transaction)
    def prepare(self, txn):
        if self.prepared:
            raise TypeError('Already prepared')
        self.prepared = True
        self._checkTransaction(txn)
        self.transaction = txn
        self.state += self.delta
        if self.objects:
            for obj in list(self.objects):
                # XXX: This _SO_obsolete check is a workaround until we get a
                # newer version of sqlos.
                #   - Andrew Bennetts, 2004-10-25
                if not obj._SO_obsolete:
                    obj.sync()
        return True

    def abort(self, txn):
        self._checkTransaction(txn)
        if self.transaction is not None:
            self.transaction = None

        self.state -= self.delta
        self.prepared = False
        self.delta = 0
        for obj in self.objects:
            expired(obj)
        self.sqlconnection.rollback()
        self.objects.clear()

    def commit(self, txn):
        if not self.prepared:
            raise TypeError('Not prepared to commit')
        self._checkTransaction(txn)
        self.state -= self.delta
        self.delta = 0
        self.transaction = None
        self.prepared = False
        for obj in self.objects:
            # XXX Can't know if modified has been fired yet. Maybe we
            # should *not* fire a modified event here?
            modified(obj)
        self.sqlconnection.commit()
        self.objects.clear()

    def register(self, obj):
        if not ISQLObject.providedBy(obj):
            raise ValueError, ("Only objects that implement ISQLObject "
                               "can be registered with this transaction "
                               "manager")
        self.objects.add(obj)
        self.delta += 1

    def savepoint(self, txn):
        """Create a savepoint that can not be rolled back"""
        if self.prepared:
            raise TypeError("Can't get savepoint during two-phase commit")
        self._checkTransaction(txn)
        self.transaction = txn
        return NoSavepointSupportRollback(self)

    def sortKey(self):
        return str(id(self))
